#include "atlas_misc.h"
#ifndef PFYDIST
   #define PFYDIST 1
#endif
#ifndef PFXDIST
   #define PFXDIST 256
#endif
#ifndef PFADIST
   #define PFADIST 256
#endif

#if !defined(ATL_3DNow) && !defined(ATL_SSE1) && \
    (PFADIST != 0 || PFXDIST != 0 || PFYDIST != 0)
   #include "atlas_prefetch.h"
#endif

#if defined(ATL_3DNow) || defined(ATL_SSE1)
   #ifndef PFIY
      #define PFIY prefetchnta
   #endif
   #ifndef PFIX
      #define PFIX prefetchnta
   #endif
   #ifndef PFIA
      #ifdef ATL_3DNow
         #define PFIA prefetchw
      #else
         #define PFIA prefetcht0
      #endif
   #endif
#endif
/*
 * X & A are prefetched in M loop PF[A,X]DIST (in bytes) ahead
 */
#if PFADIST == 0                /* flag for no prefetch */
   #define prefA(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefA(mem) __asm__ __volatile__ \
         (Mstr(PFIA) " %0" : : "m" (*(((char *)(mem))+PFADIST)))
   #else
      #if PFLVL == 2
         #define prefA(mem) ATL_pfl2W(((char*)mem)+PFADIST)
      #else
         #define prefA(mem) ATL_pfl1W(((char*)mem)+PFADIST)
      #endif
   #endif
#endif
#if PFXDIST == 0                /* flag for no prefetch */
   #define prefX(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefX(mem) __asm__ __volatile__ \
         (Mstr(PFIX) " %0" : : "m" (*(((char *)(mem))+PFXDIST)))
   #else
      #if PFLVL == 2
         #define prefX(mem) ATL_pfl2R(((char*)mem)+PFXDIST)
      #else
         #define prefX(mem) ATL_pfl1R(((char*)mem)+PFXDIST)
      #endif
   #endif
#endif
/*
 * Y is prefetched in N-loop, and always fetches next NU piece
 */
#if PFYDIST == 0                /* flag for no prefetch */
   #define prefY(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefY(mem) \
         __asm__ __volatile__ (Mstr(PFIY) " %0" : : "m" (*(((char *)(mem)))))
   #else
      #if PFLVL == 2
         #define prefY(mem) ATL_pfl2R(mem)
      #else
         #define prefY(mem) ATL_pfl1R(mem)
      #endif
   #endif
#endif
#ifndef ATL_CINT
   #define ATL_CINT const int
#endif
#ifndef ATL_INT
   #define ATL_INT int
#endif
/* Need to handle BETA=0 case by assigning y to zero outside of loop */
void ATL_UGEMV(ATL_CINT M0, ATL_CINT N, const TYPE *A, ATL_CINT lda0,
               const TYPE *X, TYPE *Y)
/*
 * 24x1 unrolled mvn_c.
 * Extracted from ATLAS/tune/blas/gemv/atlas-l2g.base
 */
{
   ATL_CINT N1=(N/1)*1, M=M0+M0, M24=((M0/24)*24)<<1,
            lda=lda0+lda0, lda1=lda*1;
   ATL_INT j;

   #ifdef BETA0
      for (j=0; j < M; j++)
         Y[j] = ATL_rzero;
   #endif
   for (j=N1; j; j -= 1, A += lda1, X += 2)
   {
      const float *A0=A;
      register float rx0=X[0], ix0=X[1];
      ATL_INT i;
      prefY(X+2+2-1);
      for (i=0; i < M24; i += 48)
      {
         const float ra0_0=A0[i+0], ia0_0=A0[i+1], ra1_0=A0[i+2], ia1_0=A0[i+3],
                     ra2_0=A0[i+4], ia2_0=A0[i+5], ra3_0=A0[i+6], ia3_0=A0[i+7],
                     ra4_0=A0[i+8], ia4_0=A0[i+9], ra5_0=A0[i+10],
                     ia5_0=A0[i+11], ra6_0=A0[i+12], ia6_0=A0[i+13],
                     ra7_0=A0[i+14], ia7_0=A0[i+15], ra8_0=A0[i+16],
                     ia8_0=A0[i+17], ra9_0=A0[i+18], ia9_0=A0[i+19],
                     ra10_0=A0[i+20], ia10_0=A0[i+21], ra11_0=A0[i+22],
                     ia11_0=A0[i+23], ra12_0=A0[i+24], ia12_0=A0[i+25],
                     ra13_0=A0[i+26], ia13_0=A0[i+27], ra14_0=A0[i+28],
                     ia14_0=A0[i+29], ra15_0=A0[i+30], ia15_0=A0[i+31],
                     ra16_0=A0[i+32], ia16_0=A0[i+33], ra17_0=A0[i+34],
                     ia17_0=A0[i+35], ra18_0=A0[i+36], ia18_0=A0[i+37],
                     ra19_0=A0[i+38], ia19_0=A0[i+39], ra20_0=A0[i+40],
                     ia20_0=A0[i+41], ra21_0=A0[i+42], ia21_0=A0[i+43],
                     ra22_0=A0[i+44], ia22_0=A0[i+45], ra23_0=A0[i+46],
                     ia23_0=A0[i+47];
         register float ry0=Y[i+0], iy0=Y[i+1], ry1=Y[i+2], iy1=Y[i+3],
                        ry2=Y[i+4], iy2=Y[i+5], ry3=Y[i+6], iy3=Y[i+7],
                        ry4=Y[i+8], iy4=Y[i+9], ry5=Y[i+10], iy5=Y[i+11],
                        ry6=Y[i+12], iy6=Y[i+13], ry7=Y[i+14], iy7=Y[i+15],
                        ry8=Y[i+16], iy8=Y[i+17], ry9=Y[i+18], iy9=Y[i+19],
                        ry10=Y[i+20], iy10=Y[i+21], ry11=Y[i+22], iy11=Y[i+23],
                        ry12=Y[i+24], iy12=Y[i+25], ry13=Y[i+26], iy13=Y[i+27],
                        ry14=Y[i+28], iy14=Y[i+29], ry15=Y[i+30], iy15=Y[i+31],
                        ry16=Y[i+32], iy16=Y[i+33], ry17=Y[i+34], iy17=Y[i+35],
                        ry18=Y[i+36], iy18=Y[i+37], ry19=Y[i+38], iy19=Y[i+39],
                        ry20=Y[i+40], iy20=Y[i+41], ry21=Y[i+42], iy21=Y[i+43],
                        ry22=Y[i+44], iy22=Y[i+45], ry23=Y[i+46], iy23=Y[i+47];

         prefY(Y);
         ry0 += ra0_0 * rx0;
         iy0 += ra0_0 * ix0;
         ry1 += ra1_0 * rx0;
         iy1 += ra1_0 * ix0;
         ry2 += ra2_0 * rx0;
         iy2 += ra2_0 * ix0;
         ry3 += ra3_0 * rx0;
         iy3 += ra3_0 * ix0;
         ry4 += ra4_0 * rx0;
         iy4 += ra4_0 * ix0;
         ry5 += ra5_0 * rx0;
         iy5 += ra5_0 * ix0;
         ry6 += ra6_0 * rx0;
         iy6 += ra6_0 * ix0;
         ry7 += ra7_0 * rx0;
         iy7 += ra7_0 * ix0;
         ry8 += ra8_0 * rx0;
         iy8 += ra8_0 * ix0;
         ry9 += ra9_0 * rx0;
         iy9 += ra9_0 * ix0;
         ry10 += ra10_0 * rx0;
         iy10 += ra10_0 * ix0;
         ry11 += ra11_0 * rx0;
         iy11 += ra11_0 * ix0;
         ry12 += ra12_0 * rx0;
         iy12 += ra12_0 * ix0;
         ry13 += ra13_0 * rx0;
         iy13 += ra13_0 * ix0;
         ry14 += ra14_0 * rx0;
         iy14 += ra14_0 * ix0;
         ry15 += ra15_0 * rx0;
         iy15 += ra15_0 * ix0;
         ry16 += ra16_0 * rx0;
         iy16 += ra16_0 * ix0;
         ry17 += ra17_0 * rx0;
         iy17 += ra17_0 * ix0;
         ry18 += ra18_0 * rx0;
         iy18 += ra18_0 * ix0;
         ry19 += ra19_0 * rx0;
         iy19 += ra19_0 * ix0;
         ry20 += ra20_0 * rx0;
         iy20 += ra20_0 * ix0;
         ry21 += ra21_0 * rx0;
         iy21 += ra21_0 * ix0;
         ry22 += ra22_0 * rx0;
         iy22 += ra22_0 * ix0;
         ry23 += ra23_0 * rx0;
         iy23 += ra23_0 * ix0;
         ry0 -= ia0_0 * ix0;
         iy0 += ia0_0 * rx0;
         prefA(A0);
         ry1 -= ia1_0 * ix0;
         iy1 += ia1_0 * rx0;
         ry2 -= ia2_0 * ix0;
         iy2 += ia2_0 * rx0;
         ry3 -= ia3_0 * ix0;
         iy3 += ia3_0 * rx0;
         ry4 -= ia4_0 * ix0;
         iy4 += ia4_0 * rx0;
         ry5 -= ia5_0 * ix0;
         iy5 += ia5_0 * rx0;
         ry6 -= ia6_0 * ix0;
         iy6 += ia6_0 * rx0;
         ry7 -= ia7_0 * ix0;
         iy7 += ia7_0 * rx0;
         ry8 -= ia8_0 * ix0;
         iy8 += ia8_0 * rx0;
         prefA(A0);
         ry9 -= ia9_0 * ix0;
         iy9 += ia9_0 * rx0;
         ry10 -= ia10_0 * ix0;
         iy10 += ia10_0 * rx0;
         ry11 -= ia11_0 * ix0;
         iy11 += ia11_0 * rx0;
         ry12 -= ia12_0 * ix0;
         iy12 += ia12_0 * rx0;
         ry13 -= ia13_0 * ix0;
         iy13 += ia13_0 * rx0;
         ry14 -= ia14_0 * ix0;
         iy14 += ia14_0 * rx0;
         ry15 -= ia15_0 * ix0;
         iy15 += ia15_0 * rx0;
         ry16 -= ia16_0 * ix0;
         iy16 += ia16_0 * rx0;
         prefA(A0);
         ry17 -= ia17_0 * ix0;
         iy17 += ia17_0 * rx0;
         ry18 -= ia18_0 * ix0;
         iy18 += ia18_0 * rx0;
         ry19 -= ia19_0 * ix0;
         iy19 += ia19_0 * rx0;
         ry20 -= ia20_0 * ix0;
         iy20 += ia20_0 * rx0;
         ry21 -= ia21_0 * ix0;
         iy21 += ia21_0 * rx0;
         ry22 -= ia22_0 * ix0;
         iy22 += ia22_0 * rx0;
         ry23 -= ia23_0 * ix0;
         iy23 += ia23_0 * rx0;


         Y[i+0] = ry0;
         Y[i+1] = iy0;
         Y[i+2] = ry1;
         Y[i+3] = iy1;
         Y[i+4] = ry2;
         Y[i+5] = iy2;
         Y[i+6] = ry3;
         Y[i+7] = iy3;
         Y[i+8] = ry4;
         Y[i+9] = iy4;
         Y[i+10] = ry5;
         Y[i+11] = iy5;
         Y[i+12] = ry6;
         Y[i+13] = iy6;
         Y[i+14] = ry7;
         Y[i+15] = iy7;
         Y[i+16] = ry8;
         Y[i+17] = iy8;
         Y[i+18] = ry9;
         Y[i+19] = iy9;
         Y[i+20] = ry10;
         Y[i+21] = iy10;
         Y[i+22] = ry11;
         Y[i+23] = iy11;
         Y[i+24] = ry12;
         Y[i+25] = iy12;
         Y[i+26] = ry13;
         Y[i+27] = iy13;
         Y[i+28] = ry14;
         Y[i+29] = iy14;
         Y[i+30] = ry15;
         Y[i+31] = iy15;
         Y[i+32] = ry16;
         Y[i+33] = iy16;
         Y[i+34] = ry17;
         Y[i+35] = iy17;
         Y[i+36] = ry18;
         Y[i+37] = iy18;
         Y[i+38] = ry19;
         Y[i+39] = iy19;
         Y[i+40] = ry20;
         Y[i+41] = iy20;
         Y[i+42] = ry21;
         Y[i+43] = iy21;
         Y[i+44] = ry22;
         Y[i+45] = iy22;
         Y[i+46] = ry23;
         Y[i+47] = iy23;
      }
      for (i=M24; i < M; i += 2)
      {
         register float ry0 = Y[i], iy0 = Y[i+1];
         const float ra0_0=A0[i], ia0_0=A0[i+1];
         ry0 += ra0_0 * rx0;
         iy0 += ra0_0 * ix0;
         ry0 -= ia0_0 * ix0;
         iy0 += ia0_0 * rx0;
         Y[i] = ry0;
         Y[i+1] = iy0;
      }
   }
}
