#include "atlas_misc.h"
#ifndef PFYDIST
   #define PFYDIST 1
#endif

#if !defined(ATL_3DNow) && !defined(ATL_SSE1) && \
    (PFADIST != 0 || PFXDIST != 0 || PFYDIST != 0)
   #include "atlas_prefetch.h"
#endif

#if defined(ATL_3DNow) || defined(ATL_SSE1)
   #ifndef PFIY
      #define PFIY prefetchnta
   #endif
   #ifndef PFIX
      #define PFIX prefetchnta
   #endif
   #ifndef PFIA
      #ifdef ATL_3DNow
         #define PFIA prefetchw
      #else
         #define PFIA prefetcht0
      #endif
   #endif
#endif
/*
 * X & A are prefetched in M loop PF[A,X]DIST (in bytes) ahead
 */
#if PFADIST == 0                /* flag for no prefetch */
   #define prefA(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefA(mem) __asm__ __volatile__ \
         (Mstr(PFIA) " %0" : : "m" (*(((char *)(mem))+PFADIST)))
   #else
      #if PFLVL == 2
         #define prefA(mem) ATL_pfl2W(((char*)mem)+PFADIST)
      #else
         #define prefA(mem) ATL_pfl1W(((char*)mem)+PFADIST)
      #endif
   #endif
#endif
#if PFXDIST == 0                /* flag for no prefetch */
   #define prefX(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefX(mem) __asm__ __volatile__ \
         (Mstr(PFIX) " %0" : : "m" (*(((char *)(mem))+PFXDIST)))
   #else
      #if PFLVL == 2
         #define prefX(mem) ATL_pfl2R(((char*)mem)+PFXDIST)
      #else
         #define prefX(mem) ATL_pfl1R(((char*)mem)+PFXDIST)
      #endif
   #endif
#endif
/*
 * Y is prefetched in N-loop, and always fetches next NU piece
 */
#if PFYDIST == 0                /* flag for no prefetch */
   #define prefY(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefY(mem) \
         __asm__ __volatile__ (Mstr(PFIY) " %0" : : "m" (*(((char *)(mem)))))
   #else
      #if PFLVL == 2
         #define prefY(mem) ATL_pfl2R(mem)
      #else
         #define prefY(mem) ATL_pfl1R(mem)
      #endif
   #endif
#endif
#ifndef ATL_CINT
   #define ATL_CINT const int
#endif
#ifndef ATL_INT
   #define ATL_INT int
#endif
void ATL_UGEMV(ATL_CINT M0, ATL_CINT N, const TYPE *A, ATL_CINT lda0,
               const TYPE *X, TYPE *Y)
/*
 * 28x1 unrolled mvt_c.
 * Extracted from ATLAS/tune/blas/gemv/atlas-l2g.base
 */
{
   ATL_CINT N1=(N/1)*1, M=M0+M0, M28=((M0/28)*28)<<1,
            lda=lda0+lda0, lda1=lda*1;
   ATL_INT j;

   for (j=N1; j; j -= 1, A += lda1, Y += 2)
   {
      const double *A0=A;
      register double ry0=ATL_rzero, iy0=ATL_rzero;
      ATL_INT i;
      prefY(Y+2+2-1);
      for (i=0; i < M28; i += 56)
      {
         const register double ra0_0=A0[i+0], ia0_0=A0[i+1], ra1_0=A0[i+2],
                               ia1_0=A0[i+3], ra2_0=A0[i+4], ia2_0=A0[i+5],
                               ra3_0=A0[i+6], ia3_0=A0[i+7], ra4_0=A0[i+8],
                               ia4_0=A0[i+9], ra5_0=A0[i+10], ia5_0=A0[i+11],
                               ra6_0=A0[i+12], ia6_0=A0[i+13], ra7_0=A0[i+14],
                               ia7_0=A0[i+15], ra8_0=A0[i+16], ia8_0=A0[i+17],
                               ra9_0=A0[i+18], ia9_0=A0[i+19], ra10_0=A0[i+20],
                               ia10_0=A0[i+21], ra11_0=A0[i+22],
                               ia11_0=A0[i+23], ra12_0=A0[i+24],
                               ia12_0=A0[i+25], ra13_0=A0[i+26],
                               ia13_0=A0[i+27], ra14_0=A0[i+28],
                               ia14_0=A0[i+29], ra15_0=A0[i+30],
                               ia15_0=A0[i+31], ra16_0=A0[i+32],
                               ia16_0=A0[i+33], ra17_0=A0[i+34],
                               ia17_0=A0[i+35], ra18_0=A0[i+36],
                               ia18_0=A0[i+37], ra19_0=A0[i+38],
                               ia19_0=A0[i+39], ra20_0=A0[i+40],
                               ia20_0=A0[i+41], ra21_0=A0[i+42],
                               ia21_0=A0[i+43], ra22_0=A0[i+44],
                               ia22_0=A0[i+45], ra23_0=A0[i+46],
                               ia23_0=A0[i+47], ra24_0=A0[i+48],
                               ia24_0=A0[i+49], ra25_0=A0[i+50],
                               ia25_0=A0[i+51], ra26_0=A0[i+52],
                               ia26_0=A0[i+53], ra27_0=A0[i+54],
                               ia27_0=A0[i+55];
         const register double rx0=X[i+0], ix0=X[i+1], rx1=X[i+2], ix1=X[i+3],
                               rx2=X[i+4], ix2=X[i+5], rx3=X[i+6], ix3=X[i+7],
                               rx4=X[i+8], ix4=X[i+9], rx5=X[i+10], ix5=X[i+11],
                               rx6=X[i+12], ix6=X[i+13], rx7=X[i+14],
                               ix7=X[i+15], rx8=X[i+16], ix8=X[i+17],
                               rx9=X[i+18], ix9=X[i+19], rx10=X[i+20],
                               ix10=X[i+21], rx11=X[i+22], ix11=X[i+23],
                               rx12=X[i+24], ix12=X[i+25], rx13=X[i+26],
                               ix13=X[i+27], rx14=X[i+28], ix14=X[i+29],
                               rx15=X[i+30], ix15=X[i+31], rx16=X[i+32],
                               ix16=X[i+33], rx17=X[i+34], ix17=X[i+35],
                               rx18=X[i+36], ix18=X[i+37], rx19=X[i+38],
                               ix19=X[i+39], rx20=X[i+40], ix20=X[i+41],
                               rx21=X[i+42], ix21=X[i+43], rx22=X[i+44],
                               ix22=X[i+45], rx23=X[i+46], ix23=X[i+47],
                               rx24=X[i+48], ix24=X[i+49], rx25=X[i+50],
                               ix25=X[i+51], rx26=X[i+52], ix26=X[i+53],
                               rx27=X[i+54], ix27=X[i+55];

         prefX(X);
         ry0 += ra0_0 * rx0;
         iy0 += ra0_0 * ix0;
         ry0 += ra1_0 * rx1;
         iy0 += ra1_0 * ix1;
         ry0 += ra2_0 * rx2;
         iy0 += ra2_0 * ix2;
         ry0 += ra3_0 * rx3;
         iy0 += ra3_0 * ix3;
         ry0 += ra4_0 * rx4;
         iy0 += ra4_0 * ix4;
         ry0 += ra5_0 * rx5;
         iy0 += ra5_0 * ix5;
         ry0 += ra6_0 * rx6;
         iy0 += ra6_0 * ix6;
         ry0 += ra7_0 * rx7;
         iy0 += ra7_0 * ix7;
         ry0 += ra8_0 * rx8;
         iy0 += ra8_0 * ix8;
         ry0 += ra9_0 * rx9;
         iy0 += ra9_0 * ix9;
         ry0 += ra10_0 * rx10;
         iy0 += ra10_0 * ix10;
         ry0 += ra11_0 * rx11;
         iy0 += ra11_0 * ix11;
         ry0 += ra12_0 * rx12;
         iy0 += ra12_0 * ix12;
         ry0 += ra13_0 * rx13;
         iy0 += ra13_0 * ix13;
         ry0 += ra14_0 * rx14;
         iy0 += ra14_0 * ix14;
         ry0 += ra15_0 * rx15;
         iy0 += ra15_0 * ix15;
         ry0 += ra16_0 * rx16;
         iy0 += ra16_0 * ix16;
         ry0 += ra17_0 * rx17;
         iy0 += ra17_0 * ix17;
         ry0 += ra18_0 * rx18;
         iy0 += ra18_0 * ix18;
         ry0 += ra19_0 * rx19;
         iy0 += ra19_0 * ix19;
         ry0 += ra20_0 * rx20;
         iy0 += ra20_0 * ix20;
         ry0 += ra21_0 * rx21;
         iy0 += ra21_0 * ix21;
         ry0 += ra22_0 * rx22;
         iy0 += ra22_0 * ix22;
         ry0 += ra23_0 * rx23;
         iy0 += ra23_0 * ix23;
         ry0 += ra24_0 * rx24;
         iy0 += ra24_0 * ix24;
         ry0 += ra25_0 * rx25;
         iy0 += ra25_0 * ix25;
         ry0 += ra26_0 * rx26;
         iy0 += ra26_0 * ix26;
         ry0 += ra27_0 * rx27;
         iy0 += ra27_0 * ix27;

         ry0 -= ia0_0 * ix0;
         iy0 += ia0_0 * rx0;
         prefA(A0);
         ry0 -= ia1_0 * ix1;
         iy0 += ia1_0 * rx1;
         ry0 -= ia2_0 * ix2;
         iy0 += ia2_0 * rx2;
         ry0 -= ia3_0 * ix3;
         iy0 += ia3_0 * rx3;
         ry0 -= ia4_0 * ix4;
         iy0 += ia4_0 * rx4;
         prefA(A0);
         ry0 -= ia5_0 * ix5;
         iy0 += ia5_0 * rx5;
         ry0 -= ia6_0 * ix6;
         iy0 += ia6_0 * rx6;
         ry0 -= ia7_0 * ix7;
         iy0 += ia7_0 * rx7;
         ry0 -= ia8_0 * ix8;
         iy0 += ia8_0 * rx8;
         prefA(A0);
         ry0 -= ia9_0 * ix9;
         iy0 += ia9_0 * rx9;
         ry0 -= ia10_0 * ix10;
         iy0 += ia10_0 * rx10;
         ry0 -= ia11_0 * ix11;
         iy0 += ia11_0 * rx11;
         ry0 -= ia12_0 * ix12;
         iy0 += ia12_0 * rx12;
         prefA(A0);
         ry0 -= ia13_0 * ix13;
         iy0 += ia13_0 * rx13;
         ry0 -= ia14_0 * ix14;
         iy0 += ia14_0 * rx14;
         ry0 -= ia15_0 * ix15;
         iy0 += ia15_0 * rx15;
         ry0 -= ia16_0 * ix16;
         iy0 += ia16_0 * rx16;
         prefA(A0);
         ry0 -= ia17_0 * ix17;
         iy0 += ia17_0 * rx17;
         ry0 -= ia18_0 * ix18;
         iy0 += ia18_0 * rx18;
         ry0 -= ia19_0 * ix19;
         iy0 += ia19_0 * rx19;
         ry0 -= ia20_0 * ix20;
         iy0 += ia20_0 * rx20;
         prefA(A0);
         ry0 -= ia21_0 * ix21;
         iy0 += ia21_0 * rx21;
         ry0 -= ia22_0 * ix22;
         iy0 += ia22_0 * rx22;
         ry0 -= ia23_0 * ix23;
         iy0 += ia23_0 * rx23;
         ry0 -= ia24_0 * ix24;
         iy0 += ia24_0 * rx24;
         prefA(A0);
         ry0 -= ia25_0 * ix25;
         iy0 += ia25_0 * rx25;
         ry0 -= ia26_0 * ix26;
         iy0 += ia26_0 * rx26;
         ry0 -= ia27_0 * ix27;
         iy0 += ia27_0 * rx27;
      }
      for (i=M28; i < M; i += 2)
      {
         const register double rx0 = X[i], ix0 = X[i+1];
         const register double ra0_0=A0[i], ia0_0=A0[i+1];
         ry0 += ra0_0 * rx0;
         iy0 += ra0_0 * ix0;
         ry0 -= ia0_0 * ix0;
         iy0 += ia0_0 * rx0;
      }
      #ifdef BETA0
         Y[0] = ry0;
         Y[1] = iy0;
      #else
         Y[0] += ry0;
         Y[1] += iy0;
      #endif
   }
}
